function [Out Lad Ctr Cf] =TMO(InColor,CGain,OffLum,windowsize,opt)   
%   [ImOut AdaptLevOut Ctr CtrStr] = TMO(InColor,CGain,OffLum,WindowSize,mode);
%
% Mario Raffin,         SIM2 Multimedia,        mraffin@sim2.it
% Gabriele Guarnieri,   University of Trieste,  gguarnieri@units.it
%   
%   INPUT PARAMETERS:
%   InColor         Input HDR Color Image
%   CGain           Contrast expansion parameter:   0 no expansion; 
%                                                   2 (typical)
%
%   OffLum          Display Luminance Offset %, (typ 1-2%)
%   WindowSize      Elaboration window size typ 21
%   mode            Mode for contrast stretch   'S' for cumulative sum (default); 
%                                               'H' for entropy 
%
%   OUTPUT PARAMETERS:
%   ImOut           Output Image
%   AdaptLevOut     Adaptation Level
%   Ctr             Evaluated Contrast
%   CtrStr          Contrast Stretch Factor
%
%   Example: 
%   [ImOut AdaptLevOut Ctr CtrStr] = TMO(memorial,2,1,21,'S');
%
%

if (nargin < 5)
    opt = 'S';
end
if (nargin < 4)
     w=21; 
else
    w=windowsize+mod(windowsize,2)-1;
end
if (nargin < 3)
    OffLum = 1;
end
if (nargin < 2)
    CGain = 2;
end
    
%0 Display characterization ------------
LdMax = 400;
Ldmin = 1;
gamma = 2.22;
%--------------------------------------


%1 INPUT IMAGE PREPARATION ------------
InColor = double(InColor);
DynDisp = (LdMax/Ldmin);

C = cmatrix('rgb2yuv', 2);
iYUV = ApplyMatrixTransform(double(InColor),C,[0 0 0]);
inY = squeeze(double(iYUV(:,:,1)));
 
% Exclude the unpopolated areas of the histogram
prcH  = prctile(inY(:), 99.9);
prcL  = prctile(inY(:),  0.1);
if (prcL == 0)
   prcL = 1e-3;
end

DynWorld = (prcH/prcL);
scale = 1/prcH ;

% 25th percentile rule
inY = inY * scale;
prc25  = prctile(inY(:), 25); 
inY(inY>1)=1;
%--------------------------------------

%2 PROCESSING --------------------------    
tic
[outY Lad Ctr Cf] = htm(inY, CGain, DynWorld, DynDisp, prc25, OffLum, w, opt);
toc
%--------------------------------------

%3 OUTPUT IMAGE PREPARATION ------------
iYUVo = iYUV;
iYUVo(:,:,2) = ((iYUV(:,:,2)./(iYUV(:,:,1)))).*outY;
iYUVo(:,:,3) = ((iYUV(:,:,3)./(iYUV(:,:,1)))).*outY;
iYUVo(:,:,1) = outY;
C = cmatrix('yuv2rgb', 2);
Out = (ApplyMatrixTransform(double(iYUVo),C,[0 0 0]));
Out = gamma709sig(Out); %move from the "linear light" domain to the 
                        %gamma compensated domain 
% --------------------------------------