/*
[Out Lad C Cf] = tmo(X, KD, W, WLum, L0w, L0d, mode)
*/

#include "mex.h"
#include "matrix.h"
#include <math.h>
#include <stdlib.h>
#include <malloc.h> 
#include <cassert>

#define sqr(x) ((x)*(x))
#define MAX(x,y) ((x >= y)?(x):(y))
#define MIN(x,y) ((x <  y)?(x):(y))

#define DF			700 //2000
#define FMARGIN     10

#define XLOGX_TABLE_SIZE      1024

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{

	// Variable declaration
	double	*Yd, *WorldLBins;
	double	*Out,*Ladapt, *C, *Cf;
	double  L0w, L0d, Cw, Cd;
	float   *fwgt, *xlogx;	// filter weights
	int     *hist;
	int		w, hw, Rows, Cols, Bins, r, c, i, j;
	int		idx_L, idx_La, bin, pixel, Cs, Ce, Rs, Re;
	int		idx, N, w2, SorH;	
	float	kd, temp1, temp2, num, den, weight, Cg;
	
	
	// Check arguments
	if ((nrhs < 6))							  //0   1  2   3    4     5     6
		mexErrMsgTxt("Usage: [Out Lad C Cf] = tmo(X, KD, W, WLum, L0w, L0d, SorH)");
	if ((nrhs < 7))		
		SorH = 0;
		
	// Check class
	if (mxGetClassID(prhs[0]) !=  mxDOUBLE_CLASS)
		mexErrMsgTxt("X must be of class DOUBLE");
	
	// Map input arguments to variables
	Yd = mxGetData(prhs[0]);
	kd = (float)mxGetScalar(prhs[1]);

	w = (int)mxGetScalar(prhs[2]);
	hw = (int)(w-1)/2;
	w2 = sqr(w);

	L0w		= (double)mxGetScalar(prhs[4]);
	L0d		= (double)mxGetScalar(prhs[5]);
	SorH	= (int)mxGetScalar(prhs[6]);
	
	WorldLBins = mxGetData(prhs[3]);
	// Measure input image
	Rows = mxGetM(prhs[0]);
	Cols = mxGetN(prhs[0]);

    //mxGetN(prhs[3]); 
	Bins = MAX(mxGetM(prhs[3]),mxGetN(prhs[3]));

	N = Rows*Cols;

	fwgt	= (float*)malloc(Bins*sizeof(float));
	xlogx	= (float*)malloc(XLOGX_TABLE_SIZE*sizeof(float));	
	hist	= (int*)malloc(Bins*sizeof(int));

	// Initialize output image and
	// Map output arguments to variables
	plhs[0] = mxCreateDoubleMatrix(Rows, Cols, mxREAL);
	Out = mxGetData(plhs[0]);
	if (nlhs>1)
	{
		plhs[1] = mxCreateDoubleMatrix(Rows, Cols, mxREAL);
		Ladapt = mxGetData(plhs[1]);
	}
	if (nlhs>2)
	{
		plhs[2] = mxCreateDoubleMatrix(Rows, Cols, mxREAL);
		C = mxGetData(plhs[2]);
	}
	if (nlhs>3)
	{
		plhs[3] = mxCreateDoubleMatrix(Rows, Cols, mxREAL);
		Cf = mxGetData(plhs[3]);
	}
			
	// Precalculate the filter weights
	temp1 = (float)((double)DF)/((float)sqr(Bins));
	for (c=0; c < Bins; c++)
	{
		weight = 1 + sqr(c)*temp1;
		fwgt[c] = 1/weight;
	} // end filter weight

	// Precalculate clogx table
	for (c=1; c <= XLOGX_TABLE_SIZE; c++)
	{
		temp1 = (float)((double)c/XLOGX_TABLE_SIZE);
		xlogx[c-1] = (float)(-1*temp1*log(temp1));		
	} 

	for (c=0; c < Cols; c++)
	{	
		Cs = c - hw;
		Ce = c + hw;
		// Condizioni al contorno sulle colonne
		if (Cs < 0)
		{
			Cs=0;
			Ce=Cs+w-1;  
		}
		else if (Ce > Cols-1)
		{
			Ce=Cols-1;
			Cs=Ce-w+1;			
		}

		// Clear histogram 
		for (bin=0; bin < Bins; bin++)
			hist[bin] = 0;
		
		for(r=0; r < Rows; r++)
		{
			Rs=r-hw;
			Re=r+hw;
			// Condizioni al contorno sulle righe
			if (Rs < 0)
			{
				Rs=0;
				Re=Rs+w-1;  
			}
			else if (Re > Rows-1)
			{
				Re=Rows-1;
				Rs=Re-w+1;				
			}

			// Histogram Calculation        
			if (r == 0) //begin the row
			{
				for (j = Cs; j <= Ce; j++)
					for (i = Rs; i <= Re; i++)
				{
					pixel = (int)(Yd[i+Rows*j]*(Bins-1));
					assert(pixel < Bins);
					hist[pixel]++;						
				}								
			}
			else if (((Rs > 0) && (Re < Rows-1)) || ((r + hw) == Rows-1))
			{
				for (j = Cs;j <= Ce; j++)
				{
					idx = Rows*j;					
					pixel = (int)(Yd[Rs-1+idx]*(Bins-1));
					assert(pixel < Bins);
					hist[pixel]--;					
					assert(idx<N);
					pixel = (int)(Yd[Re+idx]*(Bins-1));
					assert(pixel < Bins);
					hist[pixel]++;					
				}				
			}

			// Compute output adaptation level
			num = den = 0.0;
			idx_L = (int)(Yd[r+Rows*c]*(Bins-1));
			for (bin = FMARGIN; bin < (Bins-FMARGIN+1); bin++)
			{				
				temp2 = fwgt[abs(bin - idx_L)];				
				temp1 = sqr((float)hist[bin]);
				weight = temp1*temp2;
				num += weight*bin;
				den += weight;
			}

			idx_La = (int)(num/den); //adaptation level
			
			// Contrast 
			temp1 = (float)WorldLBins[idx_L];	//Lw
			temp2 = (float)WorldLBins[idx_La];	//Law			
			Cw = (double)(temp1-temp2);
			temp1 = (temp2+L0w);
			Cw /= temp1; // end contrast

			// Contrast stretch ------------------
			if (kd>0)
			{
				Cg = 0.0;
				i = MIN(idx_L,idx_La);
				j = i + abs(idx_L - idx_La);

				
				if (1 != SorH)	// Mode 0: cumulative sum 
				{
					for (bin = i; bin <= j; bin++)
					{
						Cg += (float)hist[bin];					
					}				
					Cg *= (kd/w2);
				}
				else			// Mode 1: entropy
				{										
					num = den = 0.0;
					if (j-i>0)
					{
						for (bin = i; bin <= j; bin++)
						{
							den += (float)hist[bin];					
						}
						for (bin = i; bin <= j; bin++)
						{
							if (hist[bin] != 0)
							{
								temp1	= (float)hist[bin]/den;
								temp1  *= XLOGX_TABLE_SIZE; 
								idx		= (int)(temp1);
								temp2   = temp1 - idx; //phase in [0 1[
								num += (1-temp2) * xlogx[idx-1] + temp2 * xlogx[idx];  
								//idx		= (int)(temp1 + .5);
								//num += xlogx[idx-1];							
							}						
						}
						num /= log(j-i+1); //normalization: if all the values have the same prob. the result is 1
						num = sqr(num);
					}					
					Cg = kd*num; 
				} //end Mode 1 

				Cd = Cw * (1 + Cg);
			}
			else
			{
				Cd = Cw;
				Cg = 0;
			}
			// End contrast stretch --------------

			temp1 = ((float)idx_La/Bins);//Lad (in [0 1])			
			temp2 =  Cd*(temp1 + L0d);
			
			idx = r+Rows*c;
			Out[idx]  = temp1 + temp2;

			if (nlhs>1)
				Ladapt[idx] = temp1;				
				
			if (nlhs>2)
				C[idx] = Cw;

			if (nlhs>3)
				Cf[idx] = Cg;		
		} //for r
	} // for c
	free(fwgt);
	free(hist);
	free(xlogx);
} // mexFunction